@@ -7,16 +7,19 @@ import logging
 import os
 import random
 import requests
+import time
 import psec.secrets
 import subprocess  # nosec
+import sys
 import textwrap
 
+from anytree import Node
+from anytree import RenderTree
 from bs4 import BeautifulSoup
 from cliff.command import Command
 from cliff.lister import Lister
+from collections import OrderedDict
 from configobj import ConfigObj
-from os import listdir, sep
-from os.path import abspath, basename, isdir
 from six.moves import input
 
 
@@ -29,6 +32,12 @@ AWS_CONFIG_FILE = os.path.join(os.path.expanduser('~'), '.aws', 'credentials')
 LOG = logging.getLogger(__name__)
 
 
+def remove_other_perms(dst):
+    """Make all files in path ``dst`` have ``o-rwx`` permissions."""
+    # TODO(dittrich): Test on Windows. Should work on all Linux.
+    get_output(['chmod', '-R', 'o-rwx', dst])
+
+
 def get_output(cmd=['echo', 'NO COMMAND SPECIFIED'],
                cwd=os.getcwd(),
                stderr=subprocess.STDOUT,
@@ -255,6 +264,62 @@ class MyIP(Command):
 # }
 
 
+class TfBackend(Command):
+    """
+    Enable Terraform backend support to move terraform.tfstate file
+    out of current working directory into environment path.
+    """
+
+    log = logging.getLogger(__name__)
+
+    def get_parser(self, prog_name):
+        parser = super().get_parser(prog_name)
+        parser.formatter_class = argparse.RawDescriptionHelpFormatter
+        parser.add_argument(
+            '--path',
+            action='store_true',
+            dest='path',
+            default=False,
+            help="Print path and exit (default: False)"
+        )
+        # tfstate = None
+        # try:
+        #     tfstate = os.path.join(self.app.secrets.environment_path(),
+        #                            "terraform.tfstate")
+        # except AttributeError:
+        #     pass
+        parser.epilog = textwrap.dedent("""
+            TBD(dittrich): Write this...
+            """)  # noqa
+        return parser
+
+    def take_action(self, parsed_args):
+        e = psec.secrets.SecretsEnvironment(
+                environment=self.app.options.environment)
+        tmpdir = e.tmpdir_path()
+        backend_file = os.path.join(os.getcwd(), 'tfbackend.tf')
+        tfstate_file = os.path.join(tmpdir, 'terraform.tfstate')
+        backend_text = textwrap.dedent("""\
+            terraform {{
+              backend "local" {{
+              path = "{tfstate_file}"
+              }}
+            }}
+            """.format(tfstate_file=tfstate_file))
+
+        if parsed_args.path:
+            self.log.debug('showing terraform state file path')
+            print(tfstate_file)
+        else:
+            self.log.debug('setting up terraform backend')
+            if os.path.exists(backend_file):
+                LOG.debug('updating {}'.format(backend_file))
+            else:
+                LOG.debug('creating {}'.format(backend_file))
+            with open(backend_file, 'w') as f:
+                f.write(backend_text)
+
+
 class TfOutput(Lister):
     """Retrieve current 'terraform output' results."""
 
@@ -265,7 +330,7 @@ class TfOutput(Lister):
         parser.formatter_class = argparse.RawDescriptionHelpFormatter
         tfstate = None
         try:
-            tfstate = os.path.join(self.app.secrets.environment_path(),
+            tfstate = os.path.join(self.app.secrets.tmpdir_path(),
                                    "terraform.tfstate")
         except AttributeError:
             pass
@@ -309,6 +374,12 @@ class TfOutput(Lister):
             raise RuntimeError('File does not exist: "{}"'.format(
                 tfstate)
             )
+        if self.app_args.verbose_level > 1:
+            # NOTE(dittrich): Not DRY, but spend time fixing later.
+            self.log.info(' '.join(['terraform',
+                                    'output',
+                                    '-state={}'.format(tfstate),
+                                    '-json']))
         # >> Issue: [B607:start_process_with_partial_path] Starting a process with a partial executable path  # noqa
         #    Severity: Low   Confidence: High
         #    Location: psec/utils.py:152
@@ -328,81 +399,38 @@ class TfOutput(Lister):
         return columns, data
 
 
-def tree(dir,
-         padding='',
-         print_files=True,
-         isLast=False,
-         isFirst=True,
-         outfile=None):
+def atree(dir,
+          print_files=True,
+          outfile=None):
     """
     Produces the tree structure for the path specified on the command
     line. If output is specified (e.g., as sys.stdout) it will be used,
     otherwise a list of strings is returned.
 
-    Modified code from tree.py written by Doug Dahms
-    https://stackoverflow.com/a/36253753
+    Uses anytree: https://anytree.readthedocs.io/en/latest/
 
     :param dir:
-    :param padding:
     :param print_files:
     :param outfile:
-    :param isLast:
-    :param isFirst:
     :return: str
     """
 
+    nodes = dict()
+    nodes[dir] = Node(dir)
+    root_node = nodes[dir]
+    for root, dirs, files in os.walk(dir, topdown=True):
+        if root not in nodes:
+            nodes[root] = Node(root)
+        for name in files:
+            if print_files:
+                nodes[os.path.join(root, name)] = \
+                    Node(name, parent=nodes[root])
+        for name in dirs:
+            nodes[os.path.join(root, name)] = Node(name, parent=nodes[root])
+
     output = []
-    if isFirst:
-        output.append((padding[:-1] + dir + '\n'))
-    else:
-        if isLast:
-            output.append(
-                (padding[:-1] + str('└── ') + basename(abspath(dir) + '\n')))
-        else:
-            output.append(
-                (padding[:-1] + str('├── ') + basename(abspath(dir) + '\n')))
-    files = []
-    if print_files:
-        files = listdir(dir)
-    else:
-        files = [x for x in listdir(dir) if isdir(dir + sep + x)]
-    if not isFirst:
-        padding = padding + '   '
-    files = sorted(files, key=lambda s: s.lower())
-    count = 0
-    last = len(files) - 1
-    for i, file in enumerate(files):
-        count += 1
-        path = dir + sep + file
-        isLast = (i == last)
-        if isdir(path):
-            if count == len(files):
-                if isFirst:
-                    output.extend(tree(path,
-                                       padding=padding,
-                                       print_files=print_files,
-                                       isLast=isLast,
-                                       isFirst=False,
-                                       outfile=outfile))
-                else:
-                    output.extend(tree(path,
-                                       padding=padding + ' ',
-                                       print_files=print_files,
-                                       isLast=isLast,
-                                       isFirst=False,
-                                       outfile=outfile))
-            else:
-                output.extend(tree(path,
-                                   padding=padding + '│',
-                                   print_files=print_files,
-                                   isLast=isLast,
-                                   isFirst=False,
-                                   outfile=outfile))
-        else:
-            if isLast:
-                output.append((padding + '└── ' + file + '\n'))
-            else:
-                output.append((padding + '├── ' + file + '\n'))
+    for pre, fill, node in RenderTree(root_node):
+        output.append(('{}{}'.format(pre, node.name, file=outfile)))
     if outfile is not None:
         for line in output:
             print(line, file=outfile)
@@ -410,6 +438,95 @@ def tree(dir,
         return output
 
 
+class Timer(object):
+    """
+    Timer object usable as a context manager, or for manual timing.
+    Based on code from http://coreygoldberg.blogspot.com/2012/06/python-timer-class-context-manager-for.html  # noqa
+
+    As a context manager, do:
+
+        from timer import Timer
+
+        url = 'https://github.com/timeline.json'
+
+        with Timer() as t:
+            r = requests.get(url)
+
+        print 'fetched %r in %.2f millisecs' % (url, t.elapsed*1000)
+
+    """
+
+    def __init__(self, task_description='elapsed time', verbose=False):
+        self.verbose = verbose
+        self.task_description = task_description
+        self.laps = OrderedDict()
+
+    def __enter__(self):
+        """Record initial time."""
+        self.start(lap="__enter__")
+        if self.verbose:
+            sys.stdout.write('{}...'.format(self.task_description))
+            sys.stdout.flush()
+        return self
+
+    def __exit__(self, *args):
+        """Record final time."""
+        self.stop()
+        backspace = '\b\b\b'
+        if self.verbose:
+            sys.stdout.flush()
+            if self.elapsed_raw() < 1.0:
+                sys.stdout.write(backspace + ':' + '{:.2f}ms\n'.format(
+                    self.elapsed_raw() * 1000))
+            else:
+                sys.stdout.write(backspace + ': ' + '{}\n'.format(
+                    self.elapsed()))
+            sys.stdout.flush()
+
+    def start(self, lap=None):
+        """Record starting time."""
+        t = time.time()
+        first = None if len(self.laps) == 0 \
+            else self.laps.iteritems().next()[0]
+        if first is None:
+            self.laps["__enter__"] = t
+        if lap is not None:
+            self.laps[lap] = t
+        return t
+
+    def lap(self, lap="__lap__"):
+        """
+        Records a lap time.
+        If no lap label is specified, a single 'last lap' counter will be
+        (re)used. To keep track of more laps, provide labels yourself.
+        """
+        t = time.time()
+        self.laps[lap] = t
+        return t
+
+    def stop(self):
+        """Record stop time."""
+        return self.lap(lap="__exit__")
+
+    def get_lap(self, lap="__exit__"):
+        """Get the timer for label specified by 'lap'"""
+        return self.lap[lap]
+
+    def elapsed_raw(self, start="__enter__", end="__exit__"):
+        """Return the elapsed time as a raw value."""
+        return self.laps[end] - self.laps[start]
+
+    def elapsed(self, start="__enter__", end="__exit__"):
+        """
+        Return a formatted string with elapsed time between 'start'
+        and 'end' kwargs (if specified) in HH:MM:SS.SS format.
+        """
+        hours, rem = divmod(self.elapsed_raw(start, end), 3600)
+        minutes, seconds = divmod(rem, 60)
+        return "{:0>2}:{:0>2}:{:05.2f}".format(
+            int(hours), int(minutes), seconds)
+
+
 class SetAWSCredentials(Command):
     """Set credentials from saved secrets for use by AWS CLI."""
 
